import numpy as np


def get_rotated_coords(rotation, coords):
    """Rotate coordinates by theta radians in an axis

    :param rotation: (np.ndarray) shape: (4,). [x, y, z, theta]
    :param coords: (np.ndarray) shape: (n_atoms, 3)
    :return: rotated coordinates
    """
    rot_mat = rotation_matrix(axis=rotation[:3], theta=rotation[3])

    def rotate(coord):
        return np.matmul(rot_mat, coord)

    # Map the matrix multiplication over all atomic coordinates
    return np.array(list(map(rotate, coords)))


def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.

    :param axis: (np.ndarray) Unit vector in 3D to rotate around
    :param theta: (float) Angle in radians
    """
    axis = np.asarray(axis)
    axis = axis/np.linalg.norm(axis)
    a = np.cos(theta/2.0)
    b, c, d = -axis*np.sin(theta/2.0)
    aa, bb, cc, dd = a*a, b*b, c*c, d*d
    bc, ad, ac, ab, bd, cd = b*c, a*d, a*c, a*b, b*d, c*d
    return np.array([[aa+bb-cc-dd, 2*(bc+ad), 2*(bd-ac)],
                     [2*(bc-ad), aa+cc-bb-dd, 2*(cd+ab)],
                     [2*(bd+ac), 2*(cd-ab), aa+dd-bb-cc]])


def get_points_on_sphere(n_points, r=1):
    """
    Find n evenly spaced points on a sphere using the "How to generate
    equidistributed points on the surface of a sphere" by Markus Deserno, 2004.

    Arguments:
        n_points (int): number of points to generate
        r (float): radius of the sphere

    Returns:
        (list(np.ndarray))
    """
    points = []

    a = 4.0 * np.pi * r**2 / n_points
    d = np.sqrt(a)
    m_theta = int(np.round(np.pi / d))
    d_theta = np.pi / m_theta
    d_phi = a / d_theta

    for m in range(m_theta):
        theta = np.pi * (m + 0.5) / m_theta
        m_phi = int(np.round(2.0 * np.pi * np.sin(theta)/d_phi))

        for n in range(m_phi):
            phi = 2.0 * np.pi * n / m_phi
            point = np.array([r * np.sin(theta) * np.cos(phi),
                              r * np.sin(theta) * np.sin(phi),
                              r * np.cos(theta)])

            points.append(point)

    # This algorithm doesn't always generate enough points (at least not quite)
    # so add a unit vector on the x axis to make up the difference
    if len(points) < n_points:
        points += (n_points - len(points)) * [np.array([1.0, 0.0, 0.0])]

    return points




import numpy as np
from scipy.spatial import distance_matrix
from scipy.spatial.distance import cdist
import networkx as nx
from molfunc.atoms import smiles_to_atoms
from molfunc.atoms import xyz_file_to_atoms
from molfunc.bonds import get_avg_bond_length
from molfunc.atoms import NNAtom, Atom
from molfunc.energy import get_lowest_energy_rotation
from molfunc.geom import get_rotated_coords
from molfunc.exceptions import *
from molfunc.utils import requires_atoms
from rdkit import rdBase
rdBase.DisableLog('rdApp.error')


class Molecule:

    def make_graph(self, rel_tolerance=0.2):
        """
        Make the molecular graph from the 'bonds' determined on a distance
        criteria. No distinction is made between single, double etc. bond types

        :param rel_tolerance: (float) Relative tolerance on what is classed
                                      as a bond. If a distance is
                                      < (1 + rel_tolerance) * r_avg
                                     then they are 'bonded'
        :return: None
        """

        graph = nx.Graph()
        for i in range(self.n_atoms):
            graph.add_node(i, atom_label=self.atoms[i].label)

        coordinates = self.get_coordinates()
        dist_mat = distance_matrix(coordinates, coordinates)

        # Loop over the unique pairs of atoms and add 'bonds'
        for i in range(self.n_atoms):
            for j in range(i+1, self.n_atoms):

                avg_bond_length = get_avg_bond_length(self.atoms[i].label,
                                                      self.atoms[j].label)

                # If the atoms are close enough add a bond (edge)
                if dist_mat[i, j] <= avg_bond_length * (1.0 + rel_tolerance):
                    graph.add_edge(i, j)

        self.graph = graph

        return None

    def set_atomic_valancies(self):
        """Set the atomic valency for each atom. Double/triple bonds are *not*
        distinct from single bonds"""

        for i in range(self.n_atoms):
            self.atoms[i].valence = len(list(self.graph.neighbors(i)))

        return None

    @requires_atoms()
    def translate(self, vec):
        """Translate the molecule by vector (np.ndarray, length 3)"""
        assert vec.shape == (3,)

        for atom in self.atoms:
            atom.translate(vec)
        return None

    @requires_atoms()
    def print_xyz_file(self):
        """Print a standard .xyz file from the Molecule's atoms"""

        with open(f'{self.name}.xyz', 'w') as xyz_file:
            print(self.n_atoms, '\n', file=xyz_file)
            for atom in self.atoms:
                x, y, z = atom.coord
                print(f'{atom.label:<3}{x:^10.5f}{y:^10.5f}{z:^10.5f}',
                      file=xyz_file)

        return None

    @requires_atoms()
    def get_coordinates(self):
        """Return a n_atoms x 3 shape np.ndarray containing xyz coordinates"""
        return np.array([atom.coord for atom in self.atoms])

    def set_atoms(self, atoms):
        """Set the atoms (list(molfunc.atoms.Atom)) and the number of atoms"""
        assert type(atoms) is list
        if len(atoms) > 0:
            assert isinstance(atoms[0], Atom)

        self.atoms = atoms
        self.n_atoms = len(atoms)

        return None

    def __init__(self, name='mol', xyz_filename=None, smiles=None, atoms=None):
        """
        Base molecule class. Initialised in order of priority: SMILES string,
        xyz file, atoms

        ------------------------ Keyword Arguments ----------------------------
        :param name: (str)

        :param xyz_filename: (str) .xyz filename (or filepath) from which atoms
                             will be extracted

        :param smiles: (str) SMILES string defining the molecule from which a
                       3D structure as atoms are extracted using RDKit

        :param atoms: (list(molfunc.atom.Atom)) List of atoms used to
                      initialise the molecule
        """
        self.name = str(name)
        self.n_atoms = 0
        self.graph = None
        self.atoms = None

        if smiles is not None:
            # Use RDKit to convert SMILES -> atoms
            self.set_atoms(atoms=smiles_to_atoms(smiles))

        if xyz_filename is not None:
            # Initialisation with an xyz file takes precedence over SMILES
            self.set_atoms(atoms=xyz_file_to_atoms(xyz_filename))

        if atoms is not None:
            self.set_atoms(atoms)

        if self.n_atoms != 0:
            # If there are atoms in the molecule set the graph and valancies
            self.make_graph()
            self.set_atomic_valancies()


class CoreMolecule(Molecule):

    def _check_datom_idxs(self):
        """Ensure that all atoms to be replaced by fragments are monovalent"""

        for i in self.datom_idxs:

            if not 0 <= i < self.n_atoms:
                raise DatomsNotValid(f'Can\'t functionalise an atom {i} - not '
                                     f'in the list of atoms')

            if self.atoms[i].valence == 1:
                continue

            raise DatomsNotValid(f'Cannot modify atom {self.atoms[i].label} '
                                 f'with valency {self.atoms[i].valence}')
        return None

    def get_datom_nn(self, datom_idx):
        """
        Return the nearest neighbour atom to a particular atom to delete
        (datom) along with the shift vector e.g. for a datom_idx = 1, the
        nearest neighbour is C and the vector

            vec -->

           j     i                        atoms:
           C -- H                               C, 0, 0, 0
          /                                     H, 1, 0, 0
        H                                       H, -1, 0, -1


        :param datom_idx: (int) index of the atom to delete
        :return: (molfunc.atoms.NNatom)
        """

        for (i, j) in self.graph.edges:

            vec = self.atoms[i].coord - self.atoms[j].coord

            if i == datom_idx:
                return NNAtom(atom=self.atoms[j], shift_vec=vec)
            if j == datom_idx:
                return NNAtom(atom=self.atoms[i], shift_vec=-vec)

        return None

    def _delete_datoms(self):
        """Remove all datoms from the atoms list and set the atoms"""
        return self.set_atoms(atoms=[atom for i, atom in enumerate(self.atoms)
                                     if i not in self.datom_idxs])

    def __init__(self, name='mol', xyz_filename=None, smiles=None,
                 atoms_to_del=None, atoms=None):
        """
        Core molecule class

        :param name: (str)
        :param atoms_to_del: (list(int)) List of atom indexes to delete and
                             swap for a fragment. *Indexed from 1*
        :param xyz_filename: (str)
        :param smiles: (str) SMILES string
        :param atoms: (list(molfunc.atom.Atom)) List of atoms
        """
        super().__init__(name, xyz_filename, smiles, atoms)

        # Atom indexes to delete are indexed from 1. molfunc however indexes
        # atoms from 0 so atoms_to_del are the indexes minus 1
        self.datom_idxs = [i - 1 for i in atoms_to_del] if atoms_to_del is not None else []
        self._check_datom_idxs()

        # Nearest neighbour atoms to those deleted to enable translation of the
        # fragment
        self.nn_atoms = [self.get_datom_nn(i) for i in self.datom_idxs]

        # Remove the atoms in the datom_idxs list from the atoms
        self._delete_datoms()


class FragmentMolecule(Molecule):

    def get_ratom_nearest_neighbour(self):
        """
        Return the nearest neighbour atom to the atom with label 'R' the place
        holder atom that is deleted when the Fragment and Core molecules are
        bonded. It needs to be monovalent..

        :return: (molfunc.atoms.NNatom)
        """

        # Ensure there is one R
        if not any(atom.label == 'R' for atom in self.atoms):
            raise RAtomNotFound

        # Find the first (hopefully only) monovalent R atom in the molecule
        for edge in self.graph.edges:

            for (i, j) in [edge, reversed(edge)]:
                if self.atoms[i].label != 'R':
                    continue

                if self.atoms[i].valence > 1:
                    raise RAtomInvalidValence

                return NNAtom(atom=self.atoms[j])

        # There is at least one R atom that is not bonded to any atom - return
        # the closest atom to to the R atom
        for i, atom in enumerate(self.atoms):

            if atom.label != 'R':
                continue

            distances = cdist(np.array([atom.coord]), self.get_coordinates())

            # Nearest neighbour is the second in the sorted index array, with
            # the first being the atom itself
            nn_atom_idx = np.argsort(distances)[0, 1]

            return NNAtom(atom=self.atoms[nn_atom_idx])

        raise RAtomNotFound

    def minimise_repulsion(self, other_mol):
        """
        Minimise the 'energy' with respect to rigid body rotation of this
        molecule given some other (core) molecule

        :param other_mol: (molfunc.CoreMolecule)
        """
        # Coords where the nearest neighbour atom is centered at the origin
        other_coords = other_mol.get_coordinates() - self.nn_atom.coord
        coords = self.get_coordinates() - self.nn_atom.coord

        # Set the new coordinates with the best rotation arr (lowest repulsion)
        best_r = get_lowest_energy_rotation(coords, other_coords)
        new_coords = get_rotated_coords(rotation=best_r, coords=coords)

        for i in range(self.n_atoms):
            self.atoms[i].coord = new_coords[i] + self.nn_atom.coord

        return None

    def _delete_r_atom(self):
        """Delete the atom with label 'R' from the atoms"""
        return self.set_atoms(atoms=[atom for atom in self.atoms if atom.label != 'R'])

    def _swap_fr_for_r(self):
        """Fragments can also have Fr atoms in place of R atoms, swap for R"""
        for atom in self.atoms:
            if atom.label == 'Fr':
                atom.label = 'R'

        return None

    def shift_to_core_atom(self, core_atom):
        """Given a core atom (molfunc.atoms.NNAtom) shift the fragment so the
        R atom nearest neighbour is the ideal length away from the core atom"""

        # Translate so the fragment nn atom is on top of the core atom
        self.translate(vec=core_atom.coord - self.nn_atom.coord)

        # Translate so the fragment has is the correct bond distance away
        ideal_bond_length = get_avg_bond_length(atom_i_label=core_atom.label,
                                                atom_j_label=self.nn_atom.label)
        self.translate(vec=core_atom.shift_vec * ideal_bond_length)

        # Update nn_atom coord as it will be used as the origin for rotation
        self.nn_atom.coord = core_atom.coord + core_atom.shift_vec * ideal_bond_length

        return None

    def __init__(self, name='mol', xyz_filename=None, smiles=None,
                 core_atom=None, atoms=None):
        """
        Fragment molecule class

        e.g. for a methyl self.atoms should be:
            C, 0, 0, 0
            R, 1, 0, 1
            H -1, 0, 1
            H 0, 1, -1
            H 0, -1, -1

        where the R atom will be removed and replaced for the core molecule.
        The C atom will be the nn_atom (closest to R)

        :param name: (str)

        :param core_atom: (molfunc.atoms.NNAtom) The atom on the core molecule
                          to which this fragment will be attached

        :param xyz_filename: (str)

        :param smiles: (str) SMILES string e.g. 'C[*]' or 'C[Fr]'

        :param atoms: (list(molfunc.atom.Atom)) List of atoms
        """
        super().__init__(name, xyz_filename, smiles, atoms)

        self._swap_fr_for_r()
        self.nn_atom = self.get_ratom_nearest_neighbour()
        self._delete_r_atom()

        # To manipulate as a fragment there needs to be a core atom to
        # translate to
        if core_atom is not None:
            self.shift_to_core_atom(core_atom)


class CombinedMolecule(Molecule):

    def build(self):
        """Build the combined molecule by iterating through self.fragments
        fragments minimising the repulsion to the core mol at each point"""

        if len(self.fragments) != self.n_fragments_to_add:
            raise CombinationFailed('Number of fragments is not equal to the'
                                    'number of fragments to add (core datoms)')

        atoms = self.core_mol.atoms.copy()

        for i, fragment_mol in enumerate(self.fragments):
            fragment_mol.shift_to_core_atom(self.core_mol.nn_atoms[i])

            # Repulsion is to both the core and previously added fragments
            fragment_mol.minimise_repulsion(other_mol=Molecule(atoms=atoms))

            atoms += fragment_mol.atoms

        self.set_atoms(atoms)
        return None

    def __init__(self, core_mol, frag_smiles=None, frag_smiles_list=None,
                 name='mol', fragment=None, fragments=None):
        """
        Combined molecule class

        Fragments can be added from SMILES strings (e.g. C[*]), a list of
        SMILES strings (if the core atom is functionalised more than once),
        a FragmentMolecule or a list of FragmentMolecules again if there is
        more than 1 functionalisation to be performed

        :param name: (str) Name of the molecule

        :param core_mol: (molfunc.molecules.CoreMolecule) Core molecule that
                         will be functionalised

        :param frag_smiles: (str) SMILES string to add to the core molecule
                            in *all* the core_mol.datom_idxs positions. For
                            example a methyl fragment: 'C[*]'

        :param frag_smiles_list: (list(str)) List of SMILES strings that will
                                 be added to the core molecule in sequence. The
                                 length must be equal len(core_mol.datom_idxs)

        :param fragment: (molfunc.molecules.FragmentMolecule) A pre-generated
                         fragment object. Perhaps initialised from a .xyz file
                         containing an 'R' atom which is closer than 1.5 Å
                         to a single atom

        :param fragments: (list(molfunc.molecules.FragmentMolecule)) List of
                          FragmentMolecule to add in sequence. Length of the
                          list must equal len(core_mol.datom_idxs)
        """
        super().__init__(name=name)

        self.core_mol = core_mol
        self.n_fragments_to_add = len(core_mol.datom_idxs)

        if self.n_fragments_to_add == 0:
            raise CombinationFailed('Core molecule had no datoms')

        self.fragments = []
        if frag_smiles is not None:
            self.fragments = [FragmentMolecule(smiles=frag_smiles) for _ in range(self.n_fragments_to_add)]

        if frag_smiles_list is not None:
            self.fragments = [FragmentMolecule(smiles=smiles) for smiles in frag_smiles_list]

        if fragment is not None:
            assert isinstance(fragment, FragmentMolecule)
            self.fragments = [fragment for _ in range(self.n_fragments_to_add)]

        if fragments is not None:
            assert all(isinstance(fr, FragmentMolecule) for fr in fragments)
            self.fragments = fragments

        # If there are some fragments then build the combined molecule
        if len(self.fragments) > 0:
            self.build()



import numpy as np
import os
from rdkit import Chem
from rdkit.Chem import AllChem
from molfunc.exceptions import *


class Atom:

    def __repr__(self):
        x, y, z = self.coord
        return f'[{self.label}, {x:.3f}, {y:.3f}, {z:.3f}]'

    def translate(self, vec):
        """Shift the atom in 3D space by a vector (np.ndarray) length 3"""
        self.coord += vec

    def __init__(self, atomic_symbol, x=0.0, y=0.0, z=0.0, coord=None):
        """
        Atom positioned in 3D space

        :param atomic_symbol: (str) Atomic symbol of this atom e.g. 'C' or 'Pd'
        :param x: (float) x coordinate of the atom
        :param y: (float)
        :param z: (float)

        :param coord: (np.ndarray)
        """
        self.label = str(atomic_symbol)

        if coord is not None:
            assert coord.shape == (3,)
            self.coord = coord

        else:
            self.coord = np.array([float(x), float(y), float(z)])

        # Atomic valence e.g. 4 for carbon in methane
        self.valence = 0


class NNAtom(Atom):

    def __init__(self, atom, shift_vec=None):
        """
        Nearest neighbour atom

        :param atom: (molfunc.atoms.Atom)
        :param shift_vec: (np.ndarray) length 3
        """
        super().__init__(atomic_symbol=atom.label, coord=atom.coord)
        self.shift_vec = shift_vec

        if self.shift_vec is not None:
            # Ensure the vector is normalised
            self.shift_vec = shift_vec / np.linalg.norm(shift_vec)


def xyz_file_to_atoms(filename):
    """
    From an .xyz file get a list of atoms

    :param filename: (str)
    :return: (list(molfunc.Atom))
    """
    atoms = []

    if not os.path.exists(filename):
        raise XYZfileDidNotExist

    if not filename.endswith('.xyz'):
        raise NotXYZfile

    with open(filename, 'r') as xyz_file:
        try:
            # First item in an xyz file is the number of atoms
            n_atoms = int(xyz_file.readline().split()[0])

            # XYZ lines should be the following 2 + n_atoms lines
            xyz_lines = xyz_file.readlines()[1:n_atoms+1]

        except (IndexError, ValueError):
            raise XYZfileMalformatted

        for line in xyz_lines:

            try:
                atom_label, x, y, z = line.split()[:4]
                atoms.append(Atom(atomic_symbol=atom_label, x=x, y=y, z=z))

            except (IndexError, ValueError):
                raise XYZfileMalformatted

    return atoms


def smiles_to_atoms(smiles, n_confs=1):
    """
    From a SMILES string generate n_confs conformers using RDKit with the
    ETKDGv2 algorithm

    :param smiles: (str)
    :param n_confs: (int)
    :return: (list(list(molfunc.atoms.Atom)) if n_confs > 1 else
             (list(molfunc.atoms.Atom))
    """
    # Generate an RDKit Molecule object, add hydrogens and generate conformers
    try:
        mol_obj = Chem.MolFromSmiles(smiles)
        mol_obj = Chem.AddHs(mol_obj)
        method = AllChem.ETKDGv2()
        method.pruneRmsThresh = 0.3
        conf_ids = list(AllChem.EmbedMultipleConfs(mol_obj, numConfs=n_confs,
                                                   params=method))
    except:
        # RDKit can raise unhelpful exceptions
        raise RDKitFailed

    # If n_confs > 1 then append to this list the list of Atoms
    atoms_list = []

    for i in range(len(conf_ids)):
        mol_block_lines = Chem.MolToMolBlock(mol_obj,
                                             confId=conf_ids[i]).split('\n')
        atoms = []

        # Get the atomic symbols and atomic coordinates from the block
        for line in mol_block_lines:
            items = line.split()
            if len(items) == 16:
                atom_label, x, y, z = items[3], items[0], items[1], items[2]

                atoms.append(Atom(atomic_symbol=atom_label, x=x, y=y, z=z))

        # If only one conformer is requested then only return a list of atoms
        if n_confs == 1:
            return atoms

        atoms_list.append(atoms)

    return atoms_list



import argparse
from molfunc.molecules import CoreMolecule
from molfunc.molecules import CombinedMolecule


def get_args():

    parser = argparse.ArgumentParser()
    parser.add_argument('xyz_filename', action='store', type=str,
                        help='<Required> .xyz file to functionalise. '
                             'e.g. methane.xyz')
    parser.add_argument('-a', '--atom_ids', action='store', nargs='+', type=str,
                        help='<Required> List of atom ids to swap for '
                             'fragments. e.g. 1 2')
    parser.add_argument('-s', '--fragment_smiles', action='store', type=str,
                        help='<Required> SMILES string of the fragment to add.'
                             ' e.g. [*]C')

    return parser.parse_args()


def main():

    args = get_args()
    atom_ids = [int(i) for i in args.atom_ids]

    core_mol = CoreMolecule(xyz_filename=args.xyz_filename,
                            atoms_to_del=atom_ids)

    joined = CombinedMolecule(core_mol=core_mol,
                              frag_smiles=args.fragment_smiles,
                              name=args.xyz_filename.replace('.xyz', '_mod'))
    joined.print_xyz_file()



from molfunc.geom import get_rotated_coords
from scipy.spatial import distance_matrix
from scipy.optimize import minimize
from molfunc.geom import get_points_on_sphere
import numpy as np


def repulsion(rotation, coords, alt_coords):
    """
    Calculate the repulsion energy between two sets of coordinates with

        V = Σpairs 1/rij^4

    where the coordinates are rotated in an axis x[:3] theta radians (x[3])

    :param rotation: (np.ndarray) shape: (4,)
    :param coords: (np.ndarray) shape: (n_atoms, 3)
    :param alt_coords: (np.ndarray) shape: (n_other_atoms, 3)
    """

    coords = get_rotated_coords(rotation, coords)
    dist_mat = distance_matrix(coords, alt_coords)
    energy = np.sum(np.power(dist_mat, -4))

    return energy


def get_lowest_energy_rotation(coords, other_coords, n=100):
    """
    Get the best rotation array given a set of coords, that can be rotated to
    minimise the energy with respect to another set of coordinates

    :param coords: (np.ndarray) shape (n, 3)
    :param other_coords: (np.ndarray) shape (m, 3)
    :param n: (int) Number of trials to (hopefully) find the global minimum

    :return: (np.ndarray) shape: (4,) Rotation array of an axis (length 3)
             and an angle (rad) to be used in get_rotated_coords()
    """
    # TODO parallelise
    min_energy = best_rotation = None

    # Get a set of equally distributed points on a sphere to use as axes
    axes = get_points_on_sphere(n_points=n)
    rotations = [np.append(axes[i], np.random.uniform(0.0, 2*np.pi)) for i in range(n)]

    for rotation in rotations:

        # Minimise repulsion between coords and the other_coords
        # TODO analytic derivative
        res = minimize(repulsion, x0=rotation, args=(coords, other_coords),
                       method='L-BFGS-B', tol=1.0)

        # Minimise the energy to a tighter tolerance if it's fairly low energy
        if res.fun < 5:
            res = minimize(repulsion, x0=res.x, args=(coords, other_coords),
                           method='L-BFGS-B', tol=0.05)

        if min_energy is None or res.fun < min_energy:
            min_energy = res.fun
            best_rotation = res.x

    return best_rotation



from molfunc.molecules import CombinedMolecule
from molfunc.molecules import FragmentMolecule
from molfunc.molecules import CoreMolecule


__all__ = [
    'CoreMolecule',
    'FragmentMolecule',
    'CombinedMolecule',
    ]



from setuptools import setup

setup(name='molfunc',
      version='1.0.0a2',
      packages=['molfunc'],
      license='MIT',
      author='Tom Young',
      url='https://github.com/duartegroup/molfunc',
      entry_points={'console_scripts': ['molfunc = molfunc.molfunc:main']},
      author_email='tom.young@chem.ox.ac.uk',
      description='Fast molecular functionalisation')



from functools import wraps
from molfunc.exceptions import NoAtomsInMolecule


def requires_atoms():
    """A function requiring atoms to run"""

    def func_decorator(func):
        @wraps(func)
        def wrapped_function(*args, **kwargs):
            molecule = args[0]

            assert hasattr(args[0], 'n_atoms')
            assert hasattr(args[0], 'atoms')

            if molecule.atoms is None or molecule.n_atoms == 0:
                raise NoAtomsInMolecule

            return func(*args, **kwargs)

        return wrapped_function
    return func_decorator
